/*------------------------------------------------------------------------------------------------------------
 DCF-77 Uhr mit acht 7-Segment Anzeigen  (Gem. Kathode)
 Realisiert mit ATmega16
 
 Version 1.7 vom 5.9.2012
   Messen der max. Durchlaufzeit in MAIN.entfernt. Vorzeitiges Abbrechen entfernt.

 Version 1.6 vom 5.9.2012
   Messen der max. Durchlaufzeit in MAIN. Ergab unter 2ms, ist also unkritisch.
   Daher ist das vorzeitige abbrechen der MAIN-Schleife bei gesetztem Sync-Bit berflssig.
    
 Version 1.5 vom 4.9.2012
   Sekundenzhler in 1KHz Interrupt eingebaut.
   Vorzeitiges abbrechen der MAIN-Schleife wenn Sync-Bit gesetzt ist.
   Feintuning rausgeworfen. Nur bei erfolgreicher Decodierung die Subsekunden auf 60ms setzen

 Version 1.4 vom 31.8.2012
   HS nach Zeitdecodierung auf 6 setzen
   Feintuning gendert und wieder aktiviert

 Version 1.3 vom 30.8.2012
   Bitfeld mit 2 initialisieren
   zustzliche Plausibilittsprfungen
   Zeitanpassung nicht in der Nhe der Weckzeit
   
 Version 1.2 vom 29.8.2012
   Feintuning deaktiviert
   Helligkeit via Tabelle

 Version 1.1 vom 27.8.2012
   Sommerzeitumstellung nur Sonntags
   Zeitanpassung nur von 0-4 Uhr und nur wenn alle Daten bis auf die Minute gleich sind

 Version 1.0 vom 1.5.2012
 
 
 Quarz Oszillator mit 8MHz 

          

                           ATmega 16
              Power  < PB0   1  40  PA0 <^ HHa
              Snooze < PB1   2  39  PA1 <^ HHb
         DCF Monitor < PB2   3  38  PA2 <^ MMa 
               200Hz < PB3   4  37  PA3 <^ MMb
                    ^> PB4   5  36  PA4 <^ Taste Weckzeit-aus (Stundentaste)
                    ^> PB5   6  35  PA5 <^ Taste Weckzeit-ein (Minutentaste)
     Schlummertaste ^> PB6   7  34  PA6 <^ DCF-Data
                    ^> PB7   8  33  PA7 <  LDR
                       Reset 9  32 ARef  
                       Vcc   10 31  Gnd
                       Gnd   11 30 AVcc
                       XTAL2 12 29  PC7 > Stelle Hhmm hhmm
                       XTAL1 13 28  PC6 > Stelle hHmm hhmm
          Segment a  < PD0   14 27  PC5 > Stelle hhMm hhmm
          Segment b  < PD1   15 26  PC4 > Stelle hhmM hhmm
          Segment c  < PD2   16 25  PC3 > Stelle hhmm Hhmm
          Segment d  < PD3   17 24  PC2 > Stelle hhmm hHmm
          Segment e  < PD4   18 23  PC1 > Stelle hhmm hhMm
          Segment f  < PD5   19 22  PC0 > Stelle hhmm hhmM
          Segment g  < PD6   20 21  PD7 > Segment DP
   
  ^ kennzeichnet internen Pullup 

                  
Anzeige:
  BVN-7572SR4 (gem. Kathode) rot,  14,2mm Ziffernhhe 1,7V 10mA
  DC52-11GWA  (gem. Kathode) grn, 13,2mm Ziffernhhe 2,2V 20mA  

Ausgnge:
  Kathoden: PortC (Stellen)
  Anoden:   PortD (Segmente) 
  Power           PB0 (geht an IR-Sender)
  Snooze          PB1 (geht an IR-Sender)
  DCF-Monitor:    PB2
  200Hz:          PB3 (Dient als echt volatiles Bit)

Eingnge:  
  Encoder:        PA0..PA5   HH: PA0+1  MM:PA2+3  TasterAus:PA4 TasterEin: PA5
  DCF-Empfnger:  PA6
  LDR:            PA7
  
  (AlarmAusTaste:  PB7)
  Snooze:         PB6
  
 
 Bedienung:
   Weckzeit mit den Encodern einstellen. 
   Weckzeit sichtbar mit druck auf Minutendrehgeber (Minutentaste)
   Weckzeit unsichtbar mit druck auf Stundendrehgeber (Stundentaste)
   Wecken ist aktiv, wenn Weckzeit sichtbar.
   Alarm abstellen fr 24h mit Minutentaste (grn)
   Alarm ganz aus mit Stundentaste (rot)
   Bei Alarm: 10min Schlummerpause mit Schlummertaste (schwarz)
   Irgendeine Taste >1,6s drcken wenn Radio aus: Radio ein fr 3h (aus mit Minuten- oder Stundentaste)
   
 
 
  Timer0 - Display-Multiplex (1KHz)
  Timer1 - Zeitbasis (200Hz)
  Timer2 - nicht benutzt

Der DCF-77 Empfnger ist ein externes Modul und liefert permanent ein 0V/5V Digitalsignal.

In der Startphase wird die angenommene Sekunde angezeigt 

Angezeigt wird eine interne Uhr die vom DCF-Empfang bei korrektem Empfang mintlich gestellt wird.
In der nhe der Weckzeit wird die interne Uhr nicht gestellt damit nicht versehentlich die Weckzeit 
bersprungen wird.
Ohne Empfang luft die interne Uhr mit der Zeit weg (Quarz zu ungenau). 
Das knnen schon mal 1-2 Minuten pro Tag sein.
 
Die interne Uhr wird nur gestellt, wenn die empfangenen Daten einwandfrei sind. 
Dazu werden die empfangenen Signale umfangreichen Plausibilittstests unterzogen. 
Wrde die Uhr versehentlich falsch gestellt,m ist das wesentlich schlimmer als wenn die interne Uhr 
ein paar Sekunden falsch geht, da dann ja mindestens die Minuten falsch sind. Im schlimmsten Fall
wrde die Weckzeit bersprungen.
*/


#define F_CPU 8000000UL	/* CPU clock in Hertz */

// Zum Aufnehmen der Helligkeitskurve SHOWLDR auf 1 setzen (sonst 0)
#define SHOWLDR 0   //1=Zeige PWM-Wert+Helligkeit; Einstellung PWM via Minutendrehgeber


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>  //memset

//Kurzschreibweise der Datentypen
#define U8 unsigned char
#define U16 unsigned int
#define U32 unsigned long
#define I32 signed long
#define I8 signed char
#define I16 int

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128

//AI= Analog in   DI=digital in   DO=digital out    
#define   AI_LDR 7 //  Kanal des Analogeingangs (8 Bit ADC)
#define   DI_DCF77_DATA (PINA & BIT6)

#define   DO_DCF_LED_ON       (PORTB |= BIT2)
#define   DO_DCF_LED_OFF      (PORTB &=~BIT2)
#define   DO_IR_POWER_OFF     (PORTB |= BIT0)
#define   DO_IR_POWER_ON      (PORTB &=~BIT0)
#define   DO_IR_SNOOZE_OFF    (PORTB |= BIT1)
#define   DO_IR_SNOOZE_ACTIVE (PORTB &=~BIT1)

#define   DI_SCHLUMMER_TASTE  (0==(PINB & BIT6))
#define   DI_STUNDEN_TASTE    (0==(PINA & BIT4))
#define   DI_MINUTEN_TASTE    (0==(PINA & BIT5))

#define   DO_SYNC_BIT_ON      (PORTB|= BIT3)
#define   DO_SYNC_BIT_OFF     (PORTB&=~BIT3)
#define   SYNC_BIT_IS_SET     (PORTB & BIT3)





/*Zifferncode
    111        1
   6   2    32  2
   6   2    32  2  
    777       64 
   5   3    16  4
   5   3    16  4
    444       8    
*/

#define Z0 (1+2+4+8+16+32)   
#define Z1 (2+4)   
#define Z2 (1+2+8+16+64)   
#define Z3 (1+2+4+8+64)   
#define Z4 (2+4+32+64)   
#define Z5 (1+4+8+32+64)   
#define Z6 (1+4+8+16+32+64)   
#define Z7 (1+2+4)   
#define Z8 (1+2+4+8+16+32+64)   
#define Z9 (1+2+4+8+32+64)   
#define ZE (1+8+16+32+64)   
#define ZI (4)   
#define ZN (4+16+64)   


U8 ziffern[10]={Z0,Z1,Z2,Z3,Z4,Z5,Z6,Z7,Z8,Z9};

//Helligkeitskennlinien
typedef struct{
   U8 ldr;
   U16 pwm;
}T_ldr2pwm;
T_ldr2pwm brightnessRed[]={
    {0,120},
    {7,400},
    {13,500},
    {20,900},
    {30,1200},
    {100,3000},
    {200,8170},
    {255,8170}
};    
T_ldr2pwm brightnessGreen[]={
    {0,200},
    {7,500},
    {13,800},
    {20,1400},
    {30,2000},
    {100,6000},
    {200,8170},
    {255,8170}
};    


/* DCF-77 Decoder */
/*
Whrend jeder Minute werden die Nummern von Minute, Stunde, Tag, Wochentag, 
Monat und Jahr BCD-kodiert durch Impulsmodulation der Sekundenmarken bertragen. 
Dieses "Telegramm" gilt jeweils fr die folgende Minute. 
Dabei entsprechen Sekundenmarken mit einer Dauer von 0,1 s der binren Null 
und solche mit einer Dauer von 0,2 s der binren Eins. Die Zuordnung der einzelnen 
Sekundenmarken auf die bertragene Zeitinformation zeigt das Kodierschema 


0.              Minutenbeginn (immer LOW)
1. - 14.        Reserviert, keine Bedeutung (neuedings: Verschlselte Wetterdaten)
15.             Reserveantenne aktiv
16.             Umstellung von Sommer- auf Winterzeit, oder umgekehrt
17.             Sommerzeit aktiv
18.             Winterzeit aktiv
19.             Ankndigung Schaltsekunde
20.             Zeitbeginn (Immer High)
21. - 27.       Minute 1, 2, 4, 8, 10, 20, 40
28.             Prfbit Minute
29. - 34.       Stunde 1, 2, 4, 8, 10, 20
35.             Prfbit Stunde
36. - 41.       Tag 1, 2, 4, 8, 10, 20
42. - 44.       Wochentag 1, 2, 4  (Montag=1, Sonntag=7)
45. - 49.       Monat 1, 2, 4, 8, 10
50. - 57.       Jahr 1, 2, 4, 8, 10, 20, 40, 80
58.             Prfbit Datum
59.             fehlt zur Erkennung des Minutenanfangs

(siehe auch http://www.heret.de/funkuhr/ptb.htm)

Die drei __Prfbits__ P1, P2 und P3 ergnzen jeweils die vorhergehenden Informationswrter 
(7 Bits fr die Minute, 6 Bits fr die Stunden und 22 Bits fr das Datum, einschlielich 
der Nummer des Wochentages) auf eine __gerade__ Anzahl von Einsen.


Die Zeitmarken Nr. 17 und 18 zeigen an, auf welches Zeitsystem sich die ausgesandte 
Zeitinformation bezieht. Bei Aussendung der MEZ wird die Sekundenmarke Nr. 18 auf 0,2 s
verlngert; die Sekundenmarke Nr. 17 hat eine Dauer von 0,1 s. Bei der Aussendung der 
MESZ ist es umgekehrt.

Vor einem bergang von MEZ nach MESZ oder zurck wird auerdem jeweils eine Stunde 
lang die Sekundenmarke Nr. 16 als verlngerte Marke (0,2 s) ausgesendet. Diese 
Verlngerung beginnt beim bergang von MEZ auf MESZ (von MESZ nach MEZ) um 
01.00.16 Uhr MEZ (2.00.16 Uhr MESZ) und endet um 01.59.16 Uhr MEZ (02.59.16 Uhr MESZ).

Die Sekundenmarke Nr. 19 kndigt eine Schaltsekunde an. Sie wird dann ebenfalls 
eine Stunde lang vor Einfhrung der Schaltsekunde als verlngerte Marke (0,2 s) 
ausgesendet. Schaltsekunden werden weltweit zum gleichen Zeitpunkt in die koordinierte 
Weltzeitskala UTC eingefhrt, vorzugsweise am Ende der letzen Stunde des 31. Dezember 
oder des 30. Juni. Dies bedeutet, da Schaltsekunden in der gesetzlichen Zeit der 
Bundesrepublik Deutschland eine Sekunde vor 1 Uhr MEZ am 1. Januar oder vor 
2 Uhr MESZ am 1. Juli eingeschoben werden. Bei einer Schaltsekunde am 1. Januar (1. Juli)
beginnt daher die Verlngerung der Sekundenmarke Nr. 19 um 00.00.19 Uhr MEZ (01.00.19 Uhr MESZ)
und endet um 00.59.19 Uhr MEZ (01.59.19 Uhr MESZ).

Beim Einfgen einer Schaltsekunde hat die zugehrige Minute eine Dauer von 61 Sekunden,
und die der Marke 01.00.00 Uhr MEZ (02.00.00 Uhr MESZ) vorhergehende 59. Sekundenmarke 
wird mit einer Dauer von 0,1 s ausgesendet. Die zur eingefgten 60. Sekunde gehrige 
Marke wird weggelassen (keine Trgerabsenkung).

*/



//Globale Variable
typedef struct {
  U8  bDcf77Empfang; //Wird bei Empfang jede Sekunde umgeschaltet
  U8  bBit[60];  
  U8  erfolgreichdecodiert;

  U8  uhrHH;
  U8  uhrMM;
  U8  uhrSS;
  U8  tic1s;
  U8  syncSubSec;
  U8  display[8]; //Enthlt die Bitmuster fr die 8 Anzeigen
  
  U8 weckHH;
  U8 weckMM;
  U8 alarm;        //Wird bei erreichen der Weckzeit gesetzt
  U8 weckerScharf; //Wenn >0 wird Weckzeit angezeigt
  U8 alarmdauer;   //zhlt im Minutentakt runter und macht alarm automatisch nach 1h aus
  volatile U16 OCR1BvalueRed;
  volatile U16 OCR1BvalueGreen;
  volatile U8  newOCR1B;  //sync fr 16Bit-Wert bernahme
  U8 schlummertimer;  //Minuten, nach denen erneut geweckt wird
  U8 power; //speichert momentanen Zustand des Radios
  U8 radioVergessenTimer; //Zhlt bei eingeschaltetem Power minutenweise runter. Bei 0 wird Power abgeschaltet.
  U8 keinEmpfang;

  U8 tic1ms;
}Tglob;

Tglob g;



//-----------------------------------------------------------------------------
void powerOn(void){
//-----------------------------------------------------------------------------
  DO_IR_POWER_ON;
  g.power=1;
}
//-----------------------------------------------------------------------------
void powerOff(void){
//-----------------------------------------------------------------------------
  DO_IR_POWER_OFF;
  g.power=0;
}

//-----------------------------------------------------------------------------
//Drehgeber auswerten
//  A:Richtung  B:Puls
// Bei steigender Pulsflanke Richtung auswerten
// Bei fallender Pulsflanke Richtung negiert auswerten

//--------------------------------------------------------
void drehgeberH(void){
//--------------------------------------------------------
static U8 bBisher=0;
I8 n=0;

U8 a,b;

  a=PINA & BIT0;
  b=PINA & BIT1;
  if (b!=bBisher){
    bBisher=b;
    n=g.weckHH;
    if (bBisher){ //Steigende Flanke
      if (a) n++; else n--;
    }else{
      if (a) n--; else n++;
    }
    if (n > 23) n=0;
    if (n<0) n=23;
    if (g.weckerScharf) g.weckHH=n;
  }
}

//--------------------------------------------------------
void drehgeberM(void){
//--------------------------------------------------------
static U8 bBisher=0;
I8 n=0;

U8 a,b;

  a=PINA & BIT2;
  b=PINA & BIT3;
  if (b!=bBisher){
    bBisher=b;
    n=g.weckMM;
    if (bBisher){ //steigende Flanke
      if (a) n++; else n--;
    }else{
      if (a) n--; else n++;
    }
    if (n > 59){
      n=0;
      g.weckHH++;
      if (g.weckHH>23) g.weckHH=0;
    }
    if (n<0){
      n=59;
      g.weckHH--;
      if (g.weckHH<0) g.weckHH=23;
    }
    if (g.weckerScharf) g.weckMM=n;
  }
}


//--------------------------------------------------------
U8 getAdc(U8 ch){ 
//--------------------------------------------------------
int n;
  ADMUX= (ch  & 0x07) | (1<<REFS0) | (1<<ADLAR); //Kanalwahl; Uref=AVcc; LeftAdjusted result
  asm volatile ("NOP");
  asm volatile ("NOP");
  ADCSRA=(1<<ADEN)  //Enable 
        |(1<<ADSC)  //Start
        |(1<<ADPS2)|(1<<ADPS1);  // Division Factor=64, also 125KHz ADC-Clock
  n=100;
  while ((ADCSRA & (1<<ADSC)) && n) n--; //n ist Timeout
  return ADCH; //8 Bit Ergebnis
}


//------------------------------------------------------------------------------------------------------------
ISR(TIMER0_COMP_vect){  // 1KHz
//------------------------------------------------------------------------------------------------------------
//static U8 tic=0;  
static U8 ms5=0;  
  g.tic1ms++;
  if (5==g.tic1ms){
    g.tic1ms=0;
    DO_SYNC_BIT_ON;  //alle 5ms
    if (g.syncSubSec){
      g.syncSubSec=0;
      ms5=12; //60ms nach Sekundenbeginn
    }else{
      ms5++;
      if (200==ms5){
        ms5=0;
        g.tic1s++;
      }
    }
  }
}


//--------------------------------------------------------
ISR(TIMER1_COMPA_vect){  // alle 8,192ms (122Hz) Display multiplexing
//--------------------------------------------------------
static U8  digit=0;
static U8  bitmask=1;
static U16 red, green;
  //Weiterschalten zur nchsten Stelle. Nach der letzen von vorn beginnen.
  digit++;
  digit &= 7; //Digit luft von 0..7
  if (digit)  bitmask<<=1; else bitmask=1;

  if ((digit < 3)||(7==digit)){ 
    OCR1B = red;
  }else{
    OCR1B = green;
  }

  PORTD = g.display[digit];
  PORTC = bitmask; 


  //Neue Helligkeitswerte?
  if (g.newOCR1B){
    red  = g.OCR1BvalueRed;
    green= g.OCR1BvalueGreen;
    g.newOCR1B=0; //bernahme quittieren
  }
}


//------------------------------------------------------------------------------------------------------------
ISR(TIMER1_COMPB_vect){ // Leuchtdauer)
//------------------------------------------------------------------------------------------------------------
  PORTC  = 0;  //Lowside Treiber aus -  Strom kann ber Dioden im ULN2803 abflieen
  PORTD  = 0;  //Segmente aus (Highside)
}

//------------------------------------------------------------------------------------------------------------
//ISR(TIMER2_COMP_vect){ // nicht benutzt
//}
//------------------------------------------------------------------------------------------------------------


#if 0
//--------------------------------------------------------
void displaytest(void){
//--------------------------------------------------------
U8 i,pos;

  //Alles an
  for (pos=0; pos<8; pos++) g.display[pos]=255;
  _delay_ms(1000);

  //Alle Segmente nacheinander
  for (pos=0; pos<8; pos++) g.display[pos]=1;
  for (i=0; i<8; i++){
    _delay_ms(500);
    for (pos=0; pos<8; pos++) g.display[pos]<<=1;
  }
  _delay_ms(500);
  for (pos=0; pos<8; pos++) g.display[pos]=0;

  //alle Stellen nacheinander durchlaufen mit Ziffern 0..9
  for (pos=0; pos<8; pos++){
    for (i=0; i<10; i++){
      g.display[pos] = ziffern[i];
      _delay_ms(333);
    }
    g.display[pos] = 0;
  }

  //Dezimalpunkte durchlaufen (rot hat keine)
  for (pos=4; pos<8; pos++){
    g.display[pos] = BIT7;
    _delay_ms(500);
    g.display[pos] = 0;
  }
}

#endif
//--------------------------------------------------------
void fastDisplaytest(void){
//--------------------------------------------------------
U8 i,pos;
    g.OCR1BvalueRed = 8170;
    g.OCR1BvalueGreen = 8170;

  //Alles an
  for (pos=0; pos<8; pos++){
    g.display[pos]=255;
  }
  _delay_ms(800);

  //Alle Segmente nacheinander (auf allen Stellen zugleich)
  for (pos=0; pos<8; pos++){
     g.display[pos]=1;
  }

  for (i=0; i<8; i++){
    _delay_ms(200);
    for (pos=0; pos<8; pos++){
      g.display[pos]<<=1;
    }
  }
  _delay_ms(200);
  for (pos=0; pos<8; pos++){
    g.display[pos]=0;
  }
}


//--------------------------------------------------------
U16 helligkeitRot(U8 ldr){
//--------------------------------------------------------
U8 i;
U32 deltaPwm;
U16 percent;
  if (ldr>220) return 8170;

  i=1; 
  while (ldr >= brightnessRed[i].ldr) i++;
  percent = (ldr - brightnessRed[i-1].ldr) * 100L 
          / (brightnessRed[i].ldr - brightnessRed[i-1].ldr);
  
  deltaPwm = brightnessRed[i].pwm;
  deltaPwm -= brightnessRed[i-1].pwm;
  
  return brightnessRed[i-1].pwm + (deltaPwm * percent) /100;
}

//--------------------------------------------------------
U16 helligkeitGruen(U8 ldr){
//--------------------------------------------------------
U8 i;
U32 deltaPwm;
U16 percent;
  if (ldr>220) return 8170;

  i=1; 
  while (ldr >= brightnessGreen[i].ldr) i++;
  percent = (ldr - brightnessGreen[i-1].ldr) * 100L
          / (brightnessGreen[i].ldr - brightnessGreen[i-1].ldr);
  
  deltaPwm = brightnessGreen[i].pwm;
  deltaPwm -= brightnessGreen[i-1].pwm;
  
  return brightnessGreen[i-1].pwm + (deltaPwm * percent) /100;
}



//--------------------------------------------------------
int main(void){
//--------------------------------------------------------

static I16 iTic;        //ms seit letztem Absenkungsbeginn
static U8 bSekunde=0;   //0..59

static U8 bSteigend;    //TRUE: Tiefpass bewegt sich von 0 aufwrts (Nahe Sekundenanfang)
static I8 iTiefpass;
static U8 startphase=1; 
static U8 toggle;
static U8 internTic1s=0;
static U8 hhOld,mmOld,hhNew,mmNew; //vorige/neue Minute
static U8 tag, tagOld,  wtag, wtagOld, monat,monatOld, jahr, jahrOld;
static U8 keinEmpfang;    //Zeit ohne korrekten Empfang in Minuten 
static U8 zeitumstellung; //Wird bei gesetztem DCF-Bit 16 hochgezhlt
static U8 sommerzeit;     //Wird bei gesetztem DCF-Bit 17 hochgezhlt
static U8 winterzeit;     //Wird bei gesetztem DCF-Bit 18 hochgezhlt


static I32 ldrTiefpass;
static I32 licht;

static U8 ldr=250;
static U16 eineTasteDruckzeit_ms=0;   
static U8 handEin=0;
static U8 flankenzaehler=0;
static U8 flankenJeMinute=0;
static U8 pegelBisher;
static U8 ausfallzaehler=0;
static U8 error=0;
U8 bChecksum;
U8 i;

#if SHOWLDR
U16 tmp;
#endif

  //Globale Variablen mit Null initialisieren
  memset((U8 *)&g, 0, sizeof(Tglob)); 

  DDRA  = 0;
  DDRB  = BIT0 | BIT1 | BIT2 | BIT3;
  DDRC  = BIT0 | BIT1 | BIT2 | BIT3 | BIT4 | BIT5 | BIT6 | BIT7; 
  DDRD  = BIT0 | BIT1 | BIT2 | BIT3 | BIT4 | BIT5 | BIT6 | BIT7; 

  //Pull-Up Widerstnde einschalten
  PORTA = BIT0 | BIT1 | BIT2 | BIT3 | BIT4 | BIT5 | BIT6; 
  PORTB = BIT0 | BIT1 |               BIT4 | BIT5 | BIT6 | BIT7;  
  PORTC=0;
  PORTD=0;


#if 0  //Display Leitungen durchlaufen
  n=1;
  for (i=0; i<7; i++){ 
    PORTC = n;
    _delay_ms(10);
    n<<=1;
  }
  PORTC=0;
  n=1;
  for (i=0; i<7; i++){ 
    PORTD = n;
    _delay_ms(10);
    n<<=1;
  }
  PORTD=0;
#endif

  DO_DCF_LED_ON;

  // Timer0 (1KHz. Erzeugt daraus 5ms Takt fr Uhr und DCF77)
  TCCR0 = (1<<WGM01) | (1<<CS01) | (1<<CS00);  //CTC-Modus, Clk/64  ==>125KHz 
  OCR0 = 124;         //==> 1KHz 
	TCNT0=0;

  //Timer 1 (generiert 122Hz Takt fr Display-Multiplex und 16-Bit PWM fr Helligkeit)
  TCCR1A=(1<<WGM11);
//  TCCR1B=(1<<WGM13) |(1<<WGM12) | (1<<CS10);  //CTC-Modus (Mode12), Clk/1  ==> 8MHz
  TCCR1B=(1<<WGM13) |(1<<WGM12) | (1<<CS10);  //Fast-PWM-Modus (Mode14), Clk/1  ==> 8MHz
  ICR1 = 8191;    //So weit zhlt der Timer. 8Mhz/8192/(8Stellen)=122Hz Bildfrequenz
  OCR1A = 0;
  OCR1B = 8160;  //lasse 30 Taktzyklen fr die TIMER1_COMPB-Routine 
  g.OCR1BvalueRed = OCR1B;
  g.OCR1BvalueGreen = OCR1B;
  g.newOCR1B=1;

  TIFR = 255; //Loesche alle IRQ-Bits

  //Timer2: nicht benutzt
  TCCR2 =0;

  // aktiviere Timerinterrupts 
  TIMSK = (1<<OCIE0) | (1<<OCIE1A) | (1<<OCIE1B); 

  g.weckerScharf=1; 
  powerOff();

  sei(); //Interrupts scharf
  
#if 0
  displaytest();
  _delay_ms(1000);
#else
  fastDisplaytest();
#endif

  DO_DCF_LED_OFF;
  ldrTiefpass = 200L*256L;
  drehgeberH();
  drehgeberM();
  g.weckHH=6;
  g.weckMM=29;


  while(1){  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    //Sync auf 5ms Takt
    if (!SYNC_BIT_IS_SET){ //5ms noch nicht rum. Kmmere mich um Drehgeber.
    //Encoderabfrage
      drehgeberH();
      drehgeberM();
      continue; 
    }
    DO_SYNC_BIT_OFF;

    //---hier komme ich alle 5ms hin ---


    //Eingang abfragen und filtern
    if (DI_DCF77_DATA){ //Signalabsenkung liefert High fr 100ms oder 200ms
      DO_DCF_LED_ON;
      iTiefpass++;
      if (iTiefpass>40) iTiefpass=40;  //auf 200ms begrenzen
    }else{
      DO_DCF_LED_OFF;
      iTiefpass--;
      if (iTiefpass<0){
        iTiefpass=0;
        bSteigend=1;
      }
    }
    //Pegelwechsel zaehlen um Signalqualitaet zu beurteilen
    if (pegelBisher!=DI_DCF77_DATA){
      pegelBisher=DI_DCF77_DATA;
      if(flankenzaehler<255) flankenzaehler++;
    }	    

    
    iTic+=5;  //5ms drauf
    if (iTic>3000){ //nach 3s: Kein Signal
      iTic=3000;
    }

    //Tastenabfrage
    if (DI_MINUTEN_TASTE && (0==handEin)){ //Minutentaste:Wecker einschalten - schaltet Radio aus falls grade EIN
      g.weckerScharf=1; 
      if (g.alarm || g.power){
        g.alarm =0;
        g.alarmdauer=0;
        g.radioVergessenTimer=0;
        powerOff();
      }
    }
    if (DI_STUNDEN_TASTE && (0==handEin)){
      g.weckerScharf=0;  //Stundentaste:Wecker abschalten  - schaltet Radio aus falls grade EIN
      if (g.alarm || g.power){
        g.alarm =0;
        g.alarmdauer=0;
        g.radioVergessenTimer=0;
        powerOff();
      }
    }
    
    if (DI_SCHLUMMER_TASTE && (0==handEin)){ //Wecker fr 10 Minuten aus
      if (g.alarm){
        if (g.power){
          g.schlummertimer=10;
          powerOff();
        }
      }
    }
    
    if (DI_MINUTEN_TASTE || DI_STUNDEN_TASTE || DI_SCHLUMMER_TASTE){ //Langes drcken schaltet Radio ein
      eineTasteDruckzeit_ms+=5;
      if (eineTasteDruckzeit_ms > 1600){
        eineTasteDruckzeit_ms=2000;
        handEin=100; //10*100=1000ms lang "Ein" anzeigen
        g.radioVergessenTimer=180;
        if (!g.power) powerOn();
      }
    }else{
      eineTasteDruckzeit_ms=0;
    }
    //Minutenanfang erkennen
    if (iTic > 1500){ //Minutensignal: mehr als 1s keine Absenkung
      if (59!=bSekunde){
      	flankenJeMinute = flankenzaehler;   //Bei strungsfreiem Empfang ist der Wert 118 (= 59*2)
        flankenzaehler=0;
        bSekunde=59;
        if (ausfallzaehler<2) ausfallzaehler++;
      }
    }
    else
    //Signallnge bestimmen und im Bitfeld eintragen
    if (150==iTic){ //150ms  Tiefpass ist entweder +20-10=10 oder +20+10=30
      if (iTiefpass > 19){
        g.bBit[bSekunde]=1;
      }else{
        g.bBit[bSekunde]=0;
      }
    }
    //Sekundenanfang erkennen
    if ((12==iTiefpass) && bSteigend){ //60ms nach steigender Flanke
      bSteigend=0;
      iTic=60;
      bSekunde++;
      if (58==bSekunde) ausfallzaehler=0;
      if (bSekunde>59){  //Jetzt ist Minutenanfang
        bSekunde=0;
        if (0==startphase){ //beim 1. Minutenanfang nach Einschalten ist die Zeit noch nicht bertragen
          //Auswertung zum Minutenanfang
          error=0;
          for (i=0; i<59; i++){
            if (g.bBit[i] > 1){ //Wurde mit 9 initialisiert
              error=1;  //Nicht alle Bits empfangen
            }
          }

          bChecksum=g.bBit[21]+g.bBit[22]+g.bBit[23]+g.bBit[24]+g.bBit[25]+g.bBit[26]+g.bBit[27]+g.bBit[28]; //Minute 
          if (!error && (0==(bChecksum % 2))){ //P1
            bChecksum=g.bBit[29]+g.bBit[30]+g.bBit[31]+g.bBit[32]+g.bBit[33]+g.bBit[34]+g.bBit[35];        //Stunde  
            if (0==(bChecksum % 2)){ //P2
              if (1==g.bBit[20]){ //Startbit; immer gesetzt
                bChecksum=0;
                for (i=36; i<59; i++){ //Checksumme ber Datum 
                  bChecksum+=g.bBit[i];
                }
                if (0==(bChecksum % 2)){ //P3

                  if (g.bBit[16]){
                    zeitumstellung++;
                    }else{
                    if (zeitumstellung) zeitumstellung--;
                  }
                  if (g.bBit[17]){
                    if (sommerzeit < 60) sommerzeit++;
                    }else{
                    if (sommerzeit) sommerzeit--;
                  }
                  if (g.bBit[18]){
                    if (winterzeit < 60) winterzeit++;
                    }else{
                    if (winterzeit) winterzeit--;
                  }

                  if ((118==flankenJeMinute) && (ausfallzaehler<2)){ //Nur bei ungestrtem Empfang an der Uhr drehen
                    hhOld=hhNew;
                    mmOld=mmNew;
                    tagOld = tag;
                    wtagOld=wtag;
                    monatOld=monat;
                    jahrOld=jahr;
//                    neuesDatum=0;
                    keinEmpfang=0;
                    //Decodiere Datum
                    tag = g.bBit[36]+g.bBit[37]*2+g.bBit[38]*4+g.bBit[39]*8+g.bBit[40]*10+g.bBit[41]*20;
                    wtag = g.bBit[42]+g.bBit[43]*2+g.bBit[44]*4;//1=Montag 7=Sonntag
                    monat = g.bBit[45]+g.bBit[46]*2+g.bBit[47]*4+g.bBit[48]*8;
                    jahr = g.bBit[50]+g.bBit[51]*2+g.bBit[52]*4+g.bBit[53]*8+g.bBit[54]*10+g.bBit[55]*20+g.bBit[56]*40+g.bBit[57]*80;
                    //Decodiere Zeit
                    hhNew = g.bBit[29]+g.bBit[30]*2+g.bBit[31]*4+g.bBit[32]*8+g.bBit[33]*10+g.bBit[34]*20;
                    mmNew = g.bBit[21]+g.bBit[22]*2+g.bBit[23]*4+g.bBit[24]*8+g.bBit[25]*10+g.bBit[26]*20+g.bBit[27]*40;
                    //Plausibilittstest: Neue Minute genau 1 spter als letzte Minute, alles andere gleich?
                    if (
                             (
                                (hhNew==hhOld) 
                             && (1 + mmOld == mmNew) 
//                             && ((hh < 4) //Nur nachts an der Uhr drehen 
                             && ((hhNew > g.weckHH+2)||(hhNew+2 < g.weckHH)) //nicht in Wecknhe an der Uhr drehen
                             && (jahr > 11)
                             && (wtag > 0)
                             && (monat > 0)
                             && (monat < 13)
                             && (tag > 0)
                             && (tag < 32)
                             && (tag==tagOld) 
                             && (wtag==wtagOld) 
                             && (monat==monatOld) 
                             && (jahr==jahrOld)
                             ) 
                         ||  (0==g.erfolgreichdecodiert) //Ausnahme: wenn ich noch keine Zeit kenne bin ich grozgiger
                       ){ 

                      //bernehme Zeit
                      g.syncSubSec = 1; //interne Uhr auf 60ms nach Sekundenbeginn setzen
                      g.uhrHH = hhNew;
                      g.uhrMM = mmNew;
                      g.uhrSS = 0;
                      g.erfolgreichdecodiert=1;
                    }//hhNew
                  }//118
                }//P3
              }//Bit20
            }//P2
          }//P1
        }else{// Startphase > 0
          startphase--;
        }
        for (i=0; i<60; i++){
          g.bBit[i]=9; //Initialisiere mit 9 um spter zu erkennen ob alle Bits empfangen wurden
        }
      }//(bSekunde>59)
      g.bDcf77Empfang=!g.bDcf77Empfang;
      //DO_TOGGLE_DCF_LED;
    }//Sekundenanfang (60ms)

//    if (SYNC_BIT_IS_SET) continue; //Habe jetzt keine Zeit mehr zum rechnen

    if (g.tic1s!=internTic1s){ //Sekunde rum
      internTic1s++;

      //interne Uhr eine Sekunde weiterdrehen
      g.uhrSS++;
      if (g.uhrSS>59){
        g.uhrSS=0;
        g.uhrMM++;
        if (g.alarmdauer){
          g.alarmdauer--;
          if (0==g.alarmdauer){
            g.alarm=0; //Wecker automatisch nach 1h abschalten
            g.radioVergessenTimer=0;
            g.schlummertimer=0;
            powerOff();
          }  
        }
    
        if (g.schlummertimer){
          g.schlummertimer--;
          if (!g.schlummertimer && g.alarm){
            powerOn();
            //g.schlummertimer=10;
          }
        }

        if (g.radioVergessenTimer){
          g.radioVergessenTimer--;
          if (0==g.radioVergessenTimer){
            powerOff();
          }
        }



        if (g.keinEmpfang < 100){ //Wird bei Empfang mit korreten Prfsummen auf 0 gesetzt	
          g.keinEmpfang++;
        }	
        if (g.uhrMM>59){
          g.uhrMM=0;
          g.uhrHH++;
          if (g.uhrHH>23){
            g.uhrHH=0;
            //neuesDatum=1; //Datum wird ja nicht angezeigt. Aber wenn ich das mal einbauen will, habe ich hiermit den Tagesbeginn.
          } 
          //Zeitumstellung?
          if ((zeitumstellung > 30) && (g.uhrHH >=2) && (g.uhrHH <=3) && (7==wtag)){ //"zeitumstellung" wurde in der Stunde vor der Umstellung jede Minute hochgezhlt
            if (sommerzeit > winterzeit+20){ //MESZ->MEZ
              g.uhrHH--;
            }else
            if (sommerzeit+20 < winterzeit){ //MEZ->MESZ
              g.uhrHH++;
            }
            zeitumstellung=0;
          }
        }//mm>59
        //Weckzeit erreicht?
        if ((g.uhrHH == g.weckHH) && (g.uhrMM == g.weckMM) && g.weckerScharf && g.erfolgreichdecodiert){
          g.alarm = 1;
          g.alarmdauer = 60;
          g.radioVergessenTimer=0;
          powerOn();
        }
      }//ss>59
      if (g.erfolgreichdecodiert){ //Zeit anzeigen (wenn vorhanden)
        if (0==handEin){ 
          //Zeit anzeigen
          if (g.uhrHH>9){
            g.display[0] = ziffern[g.uhrHH / 10];
          }else{
            g.display[0] = 0;
          }
          g.display[1] = ziffern[g.uhrHH % 10] | BIT7; //Dezimalpunkt einschalten (Leider hat mein rotes Display gar keinen)
          g.display[2] = ziffern[g.uhrMM / 10];
          g.display[3] = ziffern[g.uhrMM % 10];
        }
      }
    }// Sekunde rum


    toggle = !toggle;
    if (toggle){ //10ms sind rum
      if (g.weckerScharf){
        if (g.weckHH>9){
              g.display[4] = ziffern[g.weckHH / 10];
        }else{
          g.display[4] = 0;
        }
        g.display[5] = ziffern[g.weckHH % 10] | BIT7; //Dezimalpunkt einschalten (mein grnes hat einen)
        g.display[6] = ziffern[g.weckMM / 10];
        g.display[7] = ziffern[g.weckMM % 10];
      }else{
        g.display[4] = 0;
        g.display[5] = 0;
        g.display[6] = 0;
        g.display[7] = 0;
      }

      if (!g.erfolgreichdecodiert && (0==handEin)){ //DCF-Werte anzeigen
        if (DI_DCF77_DATA){
          g.display[0] = BIT0; //Strich oben
        }else{
          g.display[0] = BIT3; //Strich unten
        }
        g.display[1] = 0;    //aus
        g.display[2] = ziffern[bSekunde / 10];
        g.display[3] = ziffern[bSekunde % 10];
#if SHOWLDR  //------ LDR-Messwert in grn anzeigen -----------
        ldr = getAdc(AI_LDR); 
        g.display[4] = 0;    //aus
        g.display[5] = ziffern[ldr/ 100];
        g.display[6] = ziffern[(ldr%100) / 10];
        g.display[7] = ziffern[ldr % 10];

        //PWM mittels Minutendrehgeber einstellen und PWM-Wert in rot anzeigen
        tmp = g.weckMM;  //0..59
        tmp*=tmp;        //0..3481
        tmp*=3;          //0..10443
        if (tmp>8170) tmp=8170;
        g.display[0] = ziffern[ tmp / 1000       ];
        g.display[1] = ziffern[(tmp % 1000) / 100];
        g.display[2] = ziffern[(tmp % 100 ) /  10];
        g.display[3] = ziffern[ tmp % 10       ];
        g.OCR1BvalueRed   = tmp;
        g.OCR1BvalueGreen = tmp;
        g.newOCR1B=1;

#endif
      } //!erfolgreichdecodiert 
      if (keinEmpfang > 5)  g.display[7] |= BIT7; //Dezimalpunkt ganz rechts einschalten
      if (g.power)  g.display[4] |= BIT7; //Dezimalpunkt nach Grn-10er einschalten

    }else{//toggle
      if (handEin){//Timer noch nicht abgelaufen: Zeige " Ein" statt Uhrzeit
        handEin--;  //Timer runterzhlen
        g.display[0] = ZE; 
        g.display[1] = ZI;
        g.display[2] = ZN;
        g.display[3] = 0;
      }

      //Lichtsensor erfassen, filtern
      licht=getAdc(AI_LDR);
      ldrTiefpass+=(licht*256 - ldrTiefpass)/32;
#if 0==SHOWLDR
      if (0==g.newOCR1B){ //letzte nderung wurde bernommen
        ldr=ldrTiefpass/256;
        g.OCR1BvalueRed   = helligkeitRot(ldr);
        g.OCR1BvalueGreen = helligkeitGruen(ldr);
        g.newOCR1B=1;
      }
#endif
    }//toggle (else)

#if 0
      g.display[4] = 0;    //aus
      g.display[5] = ziffern[flankenJeMinute/ 100];
      g.display[6] = ziffern[(flankenJeMinute%100) / 10];
      g.display[7] = ziffern[flankenJeMinute % 10];
#endif
#if 0
      g.display[4] = 0;    //aus
      g.display[5] = ziffern[flankenzaehler/ 100];
      g.display[6] = ziffern[(flankenzaehler%100) / 10];
      g.display[7] = ziffern[flankenzaehler % 10];
#endif


  }//While 1
  
  return 0;
}
// EOF----------------------------------------------------------------------------------------------------------
